﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.SceneManagement;

/* Author: Andrew Brook
 * Class: IGME 202-02
 * Instructor: Jay Alan Jackson
 * Purpose of Class: Spawns and manages the various agents in the scene
 */
public class SceneController : MonoBehaviour
{
    // Public Variables
    public static SceneController instance = null;
    public GameObject walkPlane;
    public Bounds walkableBounds;
    public List<GameObject> obstacles;
    public bool drawDebugLines;
    public List<FlockAgent> flock;
    public List<Waypoint> waypointNetwork;
    public PathFollower pathFollower;
    public float pathRadius = 3.0f;
    public Vector3 averageFlockPosition;
    public Vector3 averageFlockVelocity;
    public Waypoint currentFlockWaypoint;
    public Transform averageFlockPositionTransform;
    public Material debugMaterialOrange;
    public Material debugMaterialCyan;
    // Private variables
    private Vector3 mousePosition;

    private void Awake()
    {
        // Check if an instance already exists
        if (instance == null)
        {
            // If there is no instance of a scene controller, make this game object it
            instance = this;
        }
        // Check if there is already a pre-existing instance that is not this current game object
        else if (instance != this)
        {
            // Destroy this game object
            Destroy(gameObject);
        }
        // Don't destroy the scene controller every time we load a new scene
        DontDestroyOnLoad(gameObject);
        // Calculate the viewport bounds
        walkableBounds = CalculateWalkableBounds();
    }
    // Use this for initialization
    void Start()
    {
        // Set drawDebug lines to false
        drawDebugLines = false;
        // Set the default flock waypoint
        currentFlockWaypoint = waypointNetwork[0];
    }

    // Update is called once per frame
    void Update()
    {
        if (Input.GetKeyDown(KeyCode.D))
        {
            if (drawDebugLines == false)
            {
                drawDebugLines = true;
            }
            else
            {
                drawDebugLines = false;
            }
        }
        // Calculate the average flock position and velocity
        CalculateFlockPositionAndDirection();
    }

    /// <summary>
    /// Calculates the walkable bounds of the plane
    /// </summary>
    /// <returns></returns>
    public Bounds CalculateWalkableBounds()
    {
        // Get the mesh renderer component of the walkable plane
        MeshRenderer renderer = walkPlane.GetComponent<MeshRenderer>();
        return renderer.bounds;
    }

    /// <summary>
    /// Checks for collisions between two objects using the Axis-Aligned Bounding Box method
    /// </summary>
    /// <param name="objectA"></param>
    /// <param name="objectB"></param>
    /// <returns>Whether or not a collision has occurred</returns>
    public bool AABBCollision(GameObject objectA, GameObject objectB)
    {
        // Cache the SpriteInfo components of each object
        MeshRenderer modelA = objectA.GetComponent<MeshRenderer>();
        MeshRenderer modelB = objectB.GetComponent<MeshRenderer>();

        // Check if the minimum X value of model A is greater than the maximum X value of model B
        if (modelA.bounds.min.x > modelB.bounds.max.x)
        {
            // There is not a potential collision, so return false
            return false;
        }
        // Check if the minimum X value of model B is greater than the maximum X value of model A
        else if (modelB.bounds.min.x > modelA.bounds.max.x)
        {
            // There is not a potential collision, so return false
            return false;
        }
        // Check if the minimum Z value of model A is greater than the maximum Z value of model B
        else if (modelA.bounds.min.z > modelB.bounds.max.z)
        {
            // There is not a potential collision, so return false
            return false;
        }
        // Check if the minimum Z value of model B is greater than the maximum Z value of model A
        else if (modelB.bounds.min.z > modelA.bounds.max.z)
        {
            // There is not a potential collision, so return false
            return false;
        }
        else
        {
            // There is a potential collision, so return true
            return true;
        }
    }

    /// <summary>
    /// Checks whether or not two objects are colliding using the BoundingCircle method
    /// </summary>
    /// <param name="objectA"></param>
    /// <param name="objectB"></param>
    /// <returns></returns>
    public bool CircleCollision(Vehicle vehicle, Obstacle obstacle)
    {
        // Calculate the distance between the center points of the two objects
        float distance = Vector2.Distance(vehicle.futureCenter, obstacle.center);
        // Check if the distance is less than the sum of the object's radii
        if (distance < ((vehicle.radius + obstacle.radius)) / 2)
        {
            // There is a potential collision, so return true
            return true;
        }
        else
        {
            // There is not a potential collision, so return false
            return false;
        }
    }

    /// <summary>
    /// Toggles whether or not debug lines should be drawn
    /// </summary>
    public void ToggleDrawDebugLines()
    {
        // Check if drawDebugLines is already enabled
        if (drawDebugLines == true)
        {
            // Disable draw debug lines
            drawDebugLines = false;
        }
        else
        {
            // Otherwise, drawDebugLines is disabled, so enable it
            drawDebugLines = true;
        }
    }
    /// <summary>
    /// Calculates the centroid of the flock as well as the average direction
    /// </summary>
    public void CalculateFlockPositionAndDirection()
    {
        int count = 0;
        // Reset centroid and average velocity
        averageFlockPosition = Vector3.zero;
        averageFlockVelocity = Vector3.zero;
        // Loop through each individual in the flock
        foreach (FlockAgent flockAgent in flock)
        {
            foreach (FlockAgent otherFlockAgent in flock)
            {
                // Calculate the distance from this vehicle to the flock agent in question
                float distance = Vector3.Distance(flockAgent.vehiclePosition, otherFlockAgent.vehiclePosition);
                if ((distance > 0) && (distance < flockAgent.flockNeighborDistance))
                {
                    // Increment average position
                    averageFlockPosition += flockAgent.vehiclePosition;
                    // Increment average velocity
                    averageFlockVelocity += flockAgent.velocity;
                    // Increment number of agents accounted for
                    count++;
                }
            }
        }
        // Make sure count is greater than zero
        if (count > 0)
        {
            // Average the average position
            averageFlockPosition = averageFlockPosition / count;
            averageFlockVelocity = averageFlockVelocity / count;
            averageFlockPositionTransform.position = averageFlockPosition;
            averageFlockPositionTransform.rotation = Quaternion.LookRotation(averageFlockVelocity);
        }
    }

    public void OnRenderObject()
    {
        if (drawDebugLines == true)
        {
            // Set the material to be used for the alignment line
            debugMaterialOrange.SetPass(0);
            // Draw the alignment line
            GL.Begin(GL.LINES);
            GL.Vertex(averageFlockPosition);
            GL.Vertex(averageFlockPosition + (averageFlockVelocity * 50));
            GL.End();

            // Set the material to be used for the path lines
            for (int i = 0; i < waypointNetwork.Count; i++)
            {
                debugMaterialCyan.SetPass(0);
                GL.Begin(GL.LINES);
                GL.Vertex(waypointNetwork[i].position);
                GL.Vertex(waypointNetwork[i].next.position);
                GL.End();
                /*
                if (i != waypointNetwork.Count-1)
                {
                    debugMaterialCyan.SetPass(0);
                    GL.Begin(GL.LINES);
                    GL.Vertex(waypointNetwork[i].position);
                    GL.Vertex(waypointNetwork[i].next.position);
                    GL.End();
                }
                else
                {
                    debugMaterialCyan.SetPass(0);
                    GL.Begin(GL.LINES);
                    GL.Vertex(waypointNetwork[i].position);
                    GL.Vertex(waypointNetwork[0].position);
                    GL.End();
                }
                */
            }
        }
    }
}
