﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

/* Author: Andrew Brook
 * Class: IGME 202-02
 * Instructor: Jay Alan Jackson
 * Purpose of Class: Makes an object into an autonomous agent, on which steering force behaviors will be applied
 */
public abstract class Vehicle : MonoBehaviour
{
    // Vectors necessary for force-based movement
    public Vector3 vehiclePosition;
    public Vector3 acceleration;
    public Vector3 direction;
    public Vector3 velocity;

    // Floats for force-based movement
    public float mass;
    public float frictionalCoefficient = 0.9f;
    public float maxSpeed = 1.0f;
    public float desiredSeparation;
    public bool inBounds;
    public float boundsCheckOffset = 3.0f;
    [Header("Steering Weights")]
    public float keepInBoundsWeight;
    public float obstacleAvoidanceWeight;
    public float separationWeight;
    public float seekWeight;
    public float fleeWeight;
    public float wanderWeight;
    public float cohesionWeight;
    public float alignmentWeight;
    public float pathFollowWeight;

    // Flocking Behavior Stuff
    public float flockNeighborDistance = 50.0f;
    public Vector3 averageVelocity;
    public Vector3 averagePosition;

    // Path following stuff
    public Waypoint currentWaypoint;
    public float waypointSafeDistance = 5.0f;

    // Obstacle avoidance stuff
    public float safeDistance;
    public Vector3 futureCenter
    {
        get { return vehiclePosition + (transform.forward * safeDistance); } //+ vehiclePosition + (transform.forward * safeDistance)
    }
    public float radius { get { return GetComponent<MeshRenderer>().bounds.extents.magnitude; } }
    public bool isAvoiding = false;
   
    public float wanderAngle = 0.0f;

    // Materials for debug lines
    public Material debugMaterialBlue;
    public Material debugMaterialGreen;
    public Material debugMaterialRed;
    public Material debugMaterialMagenta;
    public Material debugMaterialYellow;

    public Animator anim;
    // Private Variables
    private Vector3 obstacleAvoidanceVectorDebug = Vector3.zero;


    // Use this for initialization
    public void Start()
    {
        // Initialize vehicle position and velocity
        vehiclePosition = transform.position;
        velocity = new Vector3(0, 0, 0);
        //desiredSeparation = GetComponent<MeshRenderer>().bounds.max.x - GetComponent<MeshRenderer>().bounds.min.x;
        inBounds = true;
    }


    // Update is called once per frame
    public void Update()
    {
        // Calculate steering forces
        CalculateSteeringForces();
        // Apply friction
        //ApplyFriction();
        // Keep vehicle within bounds
        KeepInBounds();
        // Increment velocity by acceleration
        velocity += acceleration * Time.deltaTime;
        // Increment vehicle position by velocity
        vehiclePosition += velocity * Time.deltaTime;
        // Normalize velocity
        direction = velocity.normalized;
        // Rotate to direction
        RotateToDirection();
        // Zero out acceleration so we start "fresh" each frame
        acceleration = Vector3.zero;
        // Set transform position equal to vehicle position
        transform.position = vehiclePosition;

    }

    public abstract void CalculateSteeringForces();

    /// <summary>
    /// Receives an incoming force, divides it by mass, and applies it to the cumulative acceleration vector
    /// </summary>
    /// <param name="force"></param>
    public void ApplyForce(Vector3 force)
    {
        //force = new Vector3(force.x, 0, force.z);
        acceleration += force / mass;
    }


    /// <summary>
    /// Applies friction based on the frictional coefficient
    /// </summary>
    public void ApplyFriction()
    {
        // Set friction equal to the inverse of the velocity
        Vector3 friction = velocity * -1;
        // Normalize the friction
        friction.Normalize();
        // Multiply friction by the frictional coefficient
        friction = friction * frictionalCoefficient;
        // Apply the friction force
        ApplyForce(friction);
    }

    /// <summary>
    /// Keeps the vehicle within the bounds of the walkable plane
    /// </summary>
    public void KeepInBounds()
    {
        // Check if the vehicle is hitting the upper bounds of the plane
        if (vehiclePosition.z >= SceneController.instance.walkableBounds.max.z - boundsCheckOffset || vehiclePosition.z <= SceneController.instance.walkableBounds.min.z + boundsCheckOffset
            || vehiclePosition.x <= SceneController.instance.walkableBounds.min.x + boundsCheckOffset || vehiclePosition.x >= SceneController.instance.walkableBounds.max.x - boundsCheckOffset
            || vehiclePosition.y <= boundsCheckOffset || vehiclePosition.y >= SceneController.instance.gameObject.GetComponent<UnderwaterEffectControl>().waterPlane.position.y - boundsCheckOffset)
        {
            // Set inBounds to false
            inBounds = false;
            // Seek the center of the plane
            Vector3 steeringForce = Seek(SceneController.instance.walkableBounds.center) * keepInBoundsWeight;
            // Apply steering force
            ApplyForce(steeringForce * keepInBoundsWeight);
        }
        else
        {
            // Set inBounds to false
            inBounds = true;
        }
    }

    /// <summary>
    /// Calculates the force needed to steer an object towards a target, and applies it
    /// </summary>
    /// <param name="targetLocation"></param>
    public Vector3 Seek(Vector3 targetLocation)
    {
        // Calculate the desired velocity
        Vector3 desiredVelocity = targetLocation - vehiclePosition;
        // Scale the desired velocity by max speed
        desiredVelocity.Normalize();
        desiredVelocity = desiredVelocity * maxSpeed;
        // Calculate the seek force
        Vector3 seekForce = desiredVelocity - velocity;
        // Return the seek force
        return seekForce;
    }

    public Vector3 Seek(GameObject target)
    {
        return Seek(target.transform.position);
    }

    /// <summary>
    /// Calculates the force needed to steer an object away from an object, and applies it
    /// </summary>
    /// <param name="targetLocation"></param>
    public Vector3 Flee(Vector3 targetLocation)
    {
        // Calculate the desired velocity
        Vector3 desiredVelocity = targetLocation - vehiclePosition;
        // We are fleeing, so find the inverse of that desired velocity
        desiredVelocity = -desiredVelocity;
        // Scale the desired velocity by max speed
        desiredVelocity.Normalize();
        desiredVelocity = desiredVelocity * maxSpeed;
        // Calculate the flee force
        Vector3 fleeForce = desiredVelocity - velocity;
        //Return the flee force
        return fleeForce;
    }

    public Vector3 Flee(GameObject target)
    {
        return Flee(target.transform.position);
    }

    /// <summary>
    /// Rotates the vehicle to look in its current direction
    /// </summary>
    public void RotateToDirection()
    {
        // Set the rotation of the transform
        transform.forward = direction;
    }

    /// <summary>
    /// Calculates the closest obstacle to an agent, and applies a steering force to avoid it
    /// </summary>
    /// <returns></returns>
    public Vector3 ObstacleAvoidance()
    {
        // Create a new variable to represent the obstacle avoidance
        Vector3 obstacleVector = Vector3.zero;
        // Create a new variable for the closest distance and the closest object
        float minimumDistance = float.MaxValue;
        Obstacle closestObstacle = null;
        // Loop through each of the obstacles
        for (int i = 0; i < SceneController.instance.obstacles.Count; i++)
        {
            // Get the obstacle component of the obstacle
            Obstacle obstacle = SceneController.instance.obstacles[i].GetComponent<Obstacle>();
            // Get the vector between the vehicle's center and the object's center
            Vector3 vectorToCenter = obstacle.center - this.futureCenter;
            // Check if the magnitude of this vector is smaller than the current minimumDistance
            if (vectorToCenter.magnitude < minimumDistance)
            {
                // Set the minimum distance to be this new distance
                minimumDistance = vectorToCenter.magnitude;
                // Set the closest obstacle to be this obstacle
                closestObstacle = obstacle;
            }
        }
        // Start doing some calculations for the closest obstacle
        if (closestObstacle != null)
        {
            // Get the vector between the vehicle's center and the object's center again
            Vector3 vectorToCenter = closestObstacle.center - this.futureCenter;
            if (Vector3.Dot(vectorToCenter, this.transform.forward) > 0)
            {
                if (vectorToCenter.magnitude < safeDistance)
                {
                    // Get the dot product between vectorToCenter and the right vector
                    float dotProduct = Vector3.Dot(vectorToCenter, transform.right);
                    if (dotProduct > ((this.radius + closestObstacle.radius)))
                    {
                        // Set isAvoiding to true
                        if (isAvoiding != true)
                        {
                            isAvoiding = true;
                        }
                        // Set up a value for desired velocity
                        Vector3 desiredVelocity = -dotProduct * transform.right;
                        
                        // If dot product is positive, turn right
                        if (Mathf.Sign(dotProduct) == 1)
                        {
                            desiredVelocity = dotProduct * -transform.right;
                        }
                        // Otherwise, if negative, turn left
                        else
                        {
                            desiredVelocity = dotProduct * transform.right;
                        }
                        // Get the steering force
                        Vector3 steeringForce = desiredVelocity - velocity;
                        steeringForce.Normalize();
                        steeringForce *= maxSpeed;
                        obstacleAvoidanceVectorDebug = steeringForce;
                        return steeringForce;
                    }
                    else
                    {
                        if (isAvoiding != false)
                        {
                            isAvoiding = false;
                        }
                    }
                }
            }
            return Vector3.zero;
        }
        else
        {
            return Vector3.zero;
        }
    }

    /// <summary>
    /// Returns a wandering steering force for the agent, steering it towards a random position and direction
    /// </summary>
    /// <returns></returns>
    public Vector3 Wander()
    {
        // Get the vector between the vehicle's position and the future circle's center
        Vector3 vectorToCircle = futureCenter - vehiclePosition;
        // Get a random angle iteration on the unit circle
        float randomAngleIteration = Random.Range(-10.0f * Mathf.Deg2Rad, 10.0f * Mathf.Deg2Rad);
        wanderAngle += randomAngleIteration;
        // Find the point on the unit circle that this angle is associated with
        float x = futureCenter.x + Mathf.Cos(wanderAngle) * radius;
        float z = futureCenter.z + Mathf.Sin(wanderAngle) * radius;
        Vector3 targetOnCircle = new Vector3(x, transform.position.y, z);
        // Return the force resulted from seeking the target
        return Seek(targetOnCircle);
    }

    /// <summary>
    /// Aligns vehicle along an average direction calculated from the summation of its peers
    /// </summary>
    /// <returns></returns>
    public Vector3 Align()
    {
        Vector3 averageVelocity = SceneController.instance.averageFlockVelocity;
        averageVelocity.Normalize();
        averageVelocity *= maxSpeed;
        // Create a new vector for the steering force
        Vector3 steeringForce = averageVelocity - velocity;
        // Return that steering force
        return steeringForce;
    }

    /// <summary>
    /// Steers vehicle so that it groups together with its peers
    /// </summary>
    /// <returns></returns>
    public Vector3 Cohesion()
    {
        // Seek towards the centroid
        return Seek(SceneController.instance.averageFlockPosition);
    }

    /// <summary>
    /// Steers vehicle along a path based on the overarching waypoint network
    /// </summary>
    /// <returns></returns>
    public Vector3 FollowPath()
    {
        float distanceToWaypoint = Vector3.Distance(vehiclePosition, currentWaypoint.position);
        // Check if the distance is less than waypoint safe distance
        if (!(distanceToWaypoint > waypointSafeDistance))
        {
            // Vehicle has arrived at waypoint, so set to the next waypoint
            currentWaypoint = currentWaypoint.next;
        }
        // Return the seeking force to the waypoint
        return Seek(currentWaypoint.transform.position);
    }
    
    /// <summary>
    /// Called everytime the object is rendered, draws the debug lines for the zombie
    /// </summary>
    public void OnRenderObject()
    {
        // Check if drawDebugLines is enabled
        if (SceneController.instance.drawDebugLines == true)
        {
            // Set the material to be used for the forward vector line
            debugMaterialBlue.SetPass(0);
            // Draw the forward vector
            GL.Begin(GL.LINES);
            GL.Vertex(vehiclePosition);
            GL.Vertex(vehiclePosition + transform.forward);
            GL.End();
            // Set the material to be used for the right vector line
            debugMaterialGreen.SetPass(0);
            // Draw the right vector line
            GL.Begin(GL.LINES);
            GL.Vertex(vehiclePosition);
            GL.Vertex(vehiclePosition + transform.right);
            GL.End();
            // Draw the obstacle avoidance render line
            if (isAvoiding == true && obstacleAvoidanceVectorDebug != Vector3.zero)
            {
                debugMaterialRed.SetPass(0);
                GL.Begin(GL.LINES);
                GL.Vertex(vehiclePosition);
                GL.Vertex(vehiclePosition + (obstacleAvoidanceVectorDebug * 10));
                GL.End();
            }
            // Set the material to be used for the cohesion line
            debugMaterialMagenta.SetPass(0);
            // Draw the cohesion line
            GL.Begin(GL.LINES);
            GL.Vertex(vehiclePosition);
            GL.Vertex(SceneController.instance.averageFlockPosition);
            GL.End();
        }
    }
}
